;; Generated package description from company-math.el  -*- no-byte-compile: t -*-
(define-package "company-math" "1.4" "Completion backends for unicode math symbols and latex tags" '((company "0.8.0") (math-symbol-lists "1.3")) :keywords '("unicode" "symbols" "completion") :authors '(("Vitalie Spinu" . "spinuvit@gmail.com")) :maintainer '("Vitalie Spinu" . "spinuvit@gmail.com") :url "https://github.com/vspinu/company-math")
